# This file allows to configure where to save data, results, plots etc.
class CustomPaths:
    # path where downloaded data sets will be saved
    data_path = '/media/data/prog/projects/batch_active_learning'
    # path where benchmark results will be saved
    results_path = '/home/david/Dokumente/batch_active_learning/results'
    # path where plots and tables will be saved
    plots_path = 'plots'
    # path where benchmark results can be cached in a more efficient format such that they load faster
    cache_path = 'cache'


def get_data_path():
    return CustomPaths.data_path


def get_results_path():
    return CustomPaths.results_path


def get_plots_path():
    return CustomPaths.plots_path


def get_cache_path():
    return CustomPaths.cache_path